import javafx.animation.*;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.effect.Reflection;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.scene.media.Media;
import javafx.scene.media.MediaPlayer;
import javafx.util.Duration;
import logic.Song;
import logic.SongCollection;

import java.util.LinkedList;


public class MusicPlayerController {
    private SongCollection songCollection;
    private MediaPlayer mediaPlayer;
    private StringProperty currentSongURLProperty;

    @FXML private BorderPane borderPane;
    @FXML private Pane imageAndInfoWrapper;
    @FXML private StackPane bottomWrapper;
    @FXML private StackPane headingWrapper;

    @FXML private Label artistName;
    @FXML private Label songName;

    @FXML private ImageView coverImageView;
    @FXML private ImageView playPauseImageView;

    @FXML private Button playPauseButton;
    @FXML private Button songLeftButton;
    @FXML private Button songRightButton;

    private boolean mediaIsPlaying = false;
    private boolean isNextSongClicked = true;
    private final Image playImage = new Image("icons/play.png");
    private final Image pauseImage = new Image("icons/pause.png");

    private final Reflection coverReflection = new Reflection(20.0, 0.4, 0.3, 0.0);
    private final double imageWidth = 300.0;
    private final double imageHeight = 300.0;


    public void init(LinkedList<Song> songList) {
        songCollection = new SongCollection(songList);

        bindProperties();

        playCurrentSong();
    }

    private void bindProperties() {
        currentSongURLProperty = new SimpleStringProperty();
        currentSongURLProperty.bind(songCollection.getCurrentSongURLProperty());
        artistName.textProperty().bind(songCollection.getCurrentArtistProperty());
        songName.textProperty().bind(songCollection.getCurrentSongNameProperty());

        songCollection.getCurrentCoverProperty().addListener(new ChangeListener<Image>() {
            @Override
            public void changed(ObservableValue<? extends Image> observable, Image oldValue, Image newValue) {
                if (isNextSongClicked) {
                    slideNextSongCover(newValue);
                } else {
                    slidePreviousSongCover(newValue);
                }
            }
        });
    }

    private void slideNextSongCover(Image nextCover) {
        ImageView nextSlide = makeImageView(nextCover);
        imageAndInfoWrapper.getChildren().add(nextSlide);

        final ImageView oldCover = coverImageView;
        // starting position on left side
        nextSlide.relocate(-imageWidth, getYCenterForImage());

        TranslateTransition oldCoverTrans = new TranslateTransition(Duration.millis(1000), oldCover);
        // let old cover disappear
        oldCoverTrans.setByX(borderPane.getWidth() / 2 + imageWidth);
        oldCoverTrans.setCycleCount(1);
        oldCoverTrans.setOnFinished(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                imageAndInfoWrapper.getChildren().remove(oldCover);
                enableDisableButtons(false);
            }
        });

        TranslateTransition newCoverTrans = new TranslateTransition(Duration.millis(1000), nextSlide);
        newCoverTrans.setByX(getXCenterForImage() + imageWidth);
        newCoverTrans.setCycleCount(1);
        newCoverTrans.setOnFinished(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                // remove old cover
                coverImageView.setEffect(coverReflection);
                enableDisableButtons(false);
            }
        });

        coverImageView.setEffect(null);

        oldCoverTrans.play();
        newCoverTrans.play();

        coverImageView = nextSlide;
    }

    private void slidePreviousSongCover(Image nextCover) {
        ImageView nextSlide = makeImageView(nextCover);
        imageAndInfoWrapper.getChildren().add(nextSlide);

        final ImageView oldCover = coverImageView;
        // starting position on left side
        nextSlide.relocate(borderPane.getWidth(), getYCenterForImage());

        TranslateTransition oldCoverTrans = new TranslateTransition(Duration.millis(1000), oldCover);
        // let old cover disappear
        oldCoverTrans.setByX(-(borderPane.getWidth() / 2 + imageWidth));
        oldCoverTrans.setCycleCount(1);
        oldCoverTrans.setOnFinished(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                imageAndInfoWrapper.getChildren().remove(oldCover);
                enableDisableButtons(false);
            }
        });

        TranslateTransition newCoverTrans = new TranslateTransition(Duration.millis(1000), nextSlide);
        newCoverTrans.setByX(-(getXCenterForImage() + imageWidth));
        newCoverTrans.setCycleCount(1);

        newCoverTrans.setOnFinished(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                // remove old cover
                coverImageView.setEffect(coverReflection);
                enableDisableButtons(false);
            }
        });

        coverImageView.setEffect(null);

        oldCoverTrans.play();
        newCoverTrans.play();

        coverImageView = nextSlide;
    }

    private ImageView makeImageView(Image image) {
        final ImageView nextSlide = new ImageView();
        nextSlide.setFitHeight(imageHeight);
        nextSlide.setFitWidth(imageWidth);

        nextSlide.setSmooth(true);
        nextSlide.setCache(true);
        nextSlide.setPreserveRatio(true);
        nextSlide.setImage(image);

        return nextSlide;
    }

    @FXML
    public void handleLeftClick(ActionEvent actionEvent) {
        enableDisableButtons(true);

        mediaPlayer.stop();
        isNextSongClicked = false;
        songCollection.previousSong();
        playCurrentSong();
    }

    @FXML
    public void handleRightClick(ActionEvent actionEvent) {
        enableDisableButtons(true);

        mediaPlayer.stop();
        isNextSongClicked = true;
        songCollection.nextSong();
        playCurrentSong();
    }

    private void enableDisableButtons(boolean isDisabled) {
        playPauseButton.setDisable(isDisabled);
        songLeftButton.setDisable(isDisabled);
        songRightButton.setDisable(isDisabled);
    }

    @FXML
    public void handlePlayPause(ActionEvent actionEvent) {
        enableDisableButtons(true);

        if (mediaIsPlaying) {
            pauseCurrentSong();
        } else {
            playCurrentSongAfterPause();
        }

        enableDisableButtons(false);
    }

    private void playCurrentSong() {
        mediaIsPlaying = true;
        playPauseImageView.setImage(pauseImage);

        mediaPlayer = new MediaPlayer(new Media(currentSongURLProperty.getValue()));
        mediaPlayer.play();

        mediaPlayer.setOnEndOfMedia(new Runnable() {
            @Override
            public void run() {
                // play next song if end of current song
                enableDisableButtons(true);
                isNextSongClicked = true;

                songCollection.nextSong();
                playCurrentSong();
            }
        });
    }

    private void playCurrentSongAfterPause() {
        mediaIsPlaying = true;
        playPauseImageView.setImage(pauseImage);

        mediaPlayer.play();
    }

    private void pauseCurrentSong() {
        mediaIsPlaying = false;
        playPauseImageView.setImage(playImage);

        mediaPlayer.pause();
    }

    public double getMainPaneWidth() {
        return borderPane.getPrefWidth();
    }

    public double getMainPaneHeight() {
        return borderPane.getPrefHeight();
    }

    public double getXCenterForImage() {
        double currentWindowWidth = borderPane.getWidth();

        double xCenter = (currentWindowWidth / 2) - (imageWidth / 2);
        return xCenter;
    }

    public double getYCenterForImage() {
        double currentCenterHeight = borderPane.getHeight() - (headingWrapper.getHeight() + bottomWrapper.getHeight());

        double yCenter = (currentCenterHeight / 2) - ((imageHeight / 2) + 50);
        return yCenter;
    }

    public void showFirstSong() {
        coverImageView.setImage(songCollection.getCurrentCoverProperty().getValue());

        coverImageView.relocate(getXCenterForImage(), getYCenterForImage());

        relocateSongInfo(coverImageView.getX(), coverImageView.getY());
    }

    private void relocateSongInfo(double coverX, double coverY) {
        songName.relocate(getXCenterForImage(), getYCenterForImage() + imageHeight + 30);
        artistName.relocate(getXCenterForImage(), getYCenterForImage() + imageHeight + 80);
    }
}
